#ifndef LIGHT_H
#define LIGHT_H

#include <glm/vec3.hpp>
#include <glm/mat4x4.hpp>
#include <string>
#include <vector>
#include <SparrowSerializer/serializable.h>

class FrameBuffer;
class Shader;
class Scene;
class Texture;
class Camera;

class Light : public Serializable
{
public:
    enum LightType {
        AMBIENT,
        DIRECTIONNAL,
        POINT,
        SPOT
    };

    enum LightFlags {
        AMBIENT_FLAG,
        DIRECTIONNAL_FLAG,
        POINT_FLAG,
        SPOT_FLAG,
        SHADOWMAP_FLAG,
        CASCADED_FLAG,
        UNLIT_FLAG,
        NB_FLAGS
    };
    static const char* flagStr[];
    static const glm::mat4 biasMatrix;

    virtual LightType getType() = 0;

    void setColor(const glm::vec3 &color) { m_color = color; }
    glm::vec3 getColor() { return m_color; }

    virtual bool isShadowCaster() { return false; }

    virtual void bindAttributes(Shader *shader, Camera *camera) = 0;

    virtual unsigned int getFlags() = 0;

protected:
    P_VEC3(m_color)
};

class AmbientLight : public Light
{
    Texture* ambient;
    Texture* reflect;
    static Texture* brdfLUT;
public:
    SERIALIZABLE(AmbientLight, CAST(m_color))

    AmbientLight(Texture* ambientTex = nullptr, Texture* reflectTex = nullptr) : ambient(ambientTex), reflect(reflectTex) {}
    virtual LightType getType() { return AMBIENT; }
    virtual void bindAttributes(Shader *shader, Camera *camera);
    virtual unsigned int getFlags() { return 1 << AMBIENT_FLAG; }
    void setAmbient(Texture* texture) { ambient = texture; }
    Texture* getAmbient() { return ambient; }
    void setReflect(Texture* texture) { reflect = texture; }
    Texture* getReflect() { return reflect; }
    static void setBrdfLut(Texture* texture);
};

class DirectionnalLight : public Light
{
public:
    SERIALIZABLE(DirectionnalLight,
            CAST(m_color),
            CAST(m_direction),
            CAST(m_center),
            CAST(m_shadowCaster),
            CAST(m_shadowMapResolution))

    DirectionnalLight(glm::vec3 dir = glm::vec3(0, -1, 0), glm::vec3 lightColor = glm::vec3(1));

    void setDir(const glm::vec3 &dir) { m_direction = dir; }
    glm::vec3 getDir() { return m_direction; }
    
    virtual LightType getType() { return DIRECTIONNAL; }
    virtual void bindAttributes(Shader *shader, Camera *camera);
    virtual unsigned int getFlags();
    virtual bool isShadowCaster() { return m_shadowCaster; }
    
    void initShadowMap(int resolution, bool isCascaded = false);
    void destroyShadowMap();
    void setShadowView(glm::vec3 dim = glm::vec3(1), glm::vec3 center = glm::vec3(0));
    void setCascadedShadowView(Camera *camera);
    void updateShadowMap(Scene* scene);

private:
    P_VEC3(m_direction)
    
    P_VEC3(m_center)
    
    P_BOOL(m_shadowCaster)
    P_INT(m_shadowMapResolution)
    FrameBuffer* m_shadowMap;
    std::vector<glm::mat4> m_viewMatrices;
    std::vector<glm::mat4> m_projectionMatrices;
    
    static int m_shaderRefCounter;
    static Shader* m_shaders[4];
};

class PointLight : public Light
{
public:
    SERIALIZABLE(PointLight,
            CAST(m_color),
            CAST(m_position),
            CAST(m_range),
            CAST(m_shadowCaster),
            CAST(m_shadowMapResolution))
    PointLight(glm::vec3 pos = glm::vec3(0), float attenuation = 1, glm::vec3 lightColor = glm::vec3(1));

    glm::vec3 getPos() { return m_position; }
    float getAttenuation() { return m_range; }
    
    void setPos(glm::vec3 pos) { m_position = pos; updateTransform(); }
    void setRange(float range) { m_range = range; updateTransform(); }

    virtual LightType getType() { return POINT; }
    virtual void bindAttributes(Shader *shader, Camera *camera);
    virtual unsigned int getFlags();
    
    void initShadowMap(int resolution);
    void destroyShadowMap();
    void updateShadowMap(Scene* scene);
    
    void tempGetShadowMap();
    
private:
    void updateTransform();

    P_VEC3(m_position)
    P_FLOAT(m_range)
    
    P_BOOL(m_shadowCaster)
    P_INT(m_shadowMapResolution)
    FrameBuffer* m_shadowMap;
    std::vector<glm::mat4> m_shadowTransforms;
    
    static int m_shaderRefCounter;
    static Shader* m_shaders[2];
};

#endif // LIGHT_H
