#include "mesh.h"
#include <glm/ext.hpp>
#include <set>

#include "sparrowrenderer.h"
#include "material.h"
#include "buffer.h"

INIT_SERIALIZABLE(Mesh)

const char* const Mesh::flagStr[Mesh::NB_FLAGS] =
{
    "INDEXED",
    "TEXTURABLE",
    "INSTANCED",

    "MESH_3D",
    "MESH_2D",

    "TANGENT_SPACE",
    "DOUBLE_SIDED",
    "BILLBOARD",
    "WIREFRAME",
    "SHADOWED",

    "COLOR_TEXTURE",
    "ALPHA_MASK",

    "PHONG",
    "ALBEDO_TEXTURE",
    "EMISSION_TEXTURE",
    "ROUGHNESS_TEXTURE",
    "METALLIC_TEXTURE",
    "NORMAL_MAP",

    "BUMP_MAP"
};

Mesh::Mesh()
{
    m_name = "mesh";
    m_material = NULL;
    m_isDoubleSided = false;
    m_isBillboard = false;
    m_isWireframe = false;
    m_isShadowCaster = true;
    m_depth = 0;
    m_flags = 0;
    m_primitive_type = GL_TRIANGLES;
    m_vao = 0;
    clearBuffers();
}

Mesh::Mesh(const std::string &name) :
    Mesh()
{
    m_name = name;
}

Mesh::~Mesh()
{
    destroyGL();
}

void Mesh::addBuffer(Buffer *b, int bufferType)
{
    buffersId[bufferType] = buffers.size();
    buffers.push_back(b);
}

void Mesh::clearBuffers()
{
    for(int i=0; i<NB_BUFFERS; ++i)
        buffersId[i] = -1;
    for(Buffer *b : buffers)
        delete(b);
    buffers.clear();
}

unsigned int Mesh::updateFlags()
{
    m_flags = 0;

    if(!m_indices.empty())
        m_flags |= 1 << MESH_INDEXED;
    if(!m_texCoords.empty())
        m_flags |= 1 << MESH_TEXTURABLE;
    if(!m_instances_offsets.empty())
        m_flags |= 1 << MESH_INSTANCED;

    if(!m_positions3D.empty())
    {
        m_flags |= 1 << MESH_3D;

        if(!m_tangents.empty())
            m_flags |= 1 << MESH_TANGENT_SPACE;
        if(m_isDoubleSided)
            m_flags |= 1 << MESH_DOUBLE_SIDED;
        if(m_isBillboard)
            m_flags |= 1 << MESH_BILLBOARD;
        if(m_isWireframe)
            m_flags |= 1 << MESH_WIREFRAME;
        if(m_isShadowCaster)
            m_flags |= 1 << MESH_SHADOWED;
    }
    else
        m_flags |= 1 << MESH_2D;

    return m_flags | (m_material ? m_material->getFlags() : 0);
}

void Mesh::initGL()
{
    updateFlags();
    destroyGL();
    
    // create VAO
    glGenVertexArrays(1, &m_vao);
    glBindVertexArray(m_vao);

    // init positions VBO
    if(!m_positions3D.empty())
    {
        auto posBuffer = new TBuffer<glm::vec3>(m_positions3D, Buffer::VBO);
        posBuffer->setVertexAttrib(0, 3);
        addBuffer(posBuffer, POSITION3D_BUFFER);
        
        // init normals vbo
        if(!m_normals.empty())
        {
            auto normalsBuffer = new TBuffer<glm::vec3>(m_normals, Buffer::VBO);
            normalsBuffer->setVertexAttrib(1, 3);
            addBuffer(normalsBuffer, NORMAL_BUFFER);
        }
        // init tangents vbo
        if(!m_tangents.empty())
        {
            auto tangentsBuffer = new TBuffer<Tangents>(m_tangents, Buffer::VBO);
            tangentsBuffer->setVertexAttrib(3, 3, 0);
            tangentsBuffer->setVertexAttrib(4, 3, sizeof(glm::vec3));
            addBuffer(tangentsBuffer, TANGENT_BUFFER);
        }
    }
    else if(!m_positions2D.empty())
    {
        auto posBuffer = new TBuffer<glm::vec2>(m_positions2D, Buffer::VBO);
        posBuffer->setVertexAttrib(0, 2);
        addBuffer(posBuffer, POSITION2D_BUFFER);
    }
    else
    {
        fprintf(stderr, "ERROR : Mesh can't be initialized without position data");
        return;
    }
    // init texCoords vbo
    if(!m_texCoords.empty())
    {
        auto texCoordBuffer = new TBuffer<glm::vec2>(m_texCoords, Buffer::VBO);
        texCoordBuffer->setVertexAttrib(2, 2);
        addBuffer(texCoordBuffer, TEXCOORD_BUFFER);
    }
    
    // init instances vbo
    if(!m_instances_offsets.empty())
    {
        auto instancesBuffer = new TBuffer<glm::vec3>(m_instances_offsets, Buffer::VBO);
        instancesBuffer->setVertexAttrib(5, 3, 0, 1);
        addBuffer(instancesBuffer, INSTANCE_BUFFER);
    }
        
        // init EBO
    if(!m_indices.empty())
    {
        auto indicesBuffer = new TBuffer<GLuint>(m_indices, Buffer::EBO);
        addBuffer(indicesBuffer, INDICES_BUFFER);
    }

    // unbind vao
    glBindVertexArray(0);
}

void Mesh::draw(Shader* shader)
{
    if(m_isWireframe)
        glStencilOp(GL_KEEP, GL_KEEP, GL_ZERO);
    else
        glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
    if(m_isDoubleSided)
        glDisable(GL_CULL_FACE);
    if(m_material != NULL)
        m_material->bindAttributes(shader);
    glBindVertexArray(m_vao);
    
    if(m_indices.empty())
    {
        int size = m_positions3D.empty() ? m_positions2D.size() : m_positions3D.size();
        if(!m_instances_offsets.empty())
        {
            glDrawArraysInstanced(m_primitive_type, 0, size, m_instances_offsets.size());
        }
        else
        {
            glDrawArrays(m_primitive_type, 0, size);
        }
    }
    else
    {
        Buffer *b = buffers[buffersId[INDICES_BUFFER]];
        b->bind();
        if(!m_instances_offsets.empty())
        {
            glDrawElementsInstanced(m_primitive_type, m_indices.size(), GL_UNSIGNED_INT, NULL, m_instances_offsets.size());
        }
        else
        {
            glDrawElements(m_primitive_type, m_indices.size(), GL_UNSIGNED_INT, NULL);
        }
        b->unbind();
    }
    
    glBindVertexArray(0);
    if(m_isDoubleSided)
        glEnable(GL_CULL_FACE);
}

void Mesh::destroyGL()
{
    if(m_vao != 0)
    {
        glDeleteVertexArrays(1, &m_vao);
        m_vao = 0;
    }
    clearBuffers();
}

void Mesh::clearData()
{
    m_positions3D.clear();
    m_normals.clear();
    m_tangents.clear();
    m_positions2D.clear();
    m_texCoords.clear();
    m_instances_offsets.clear();
    m_indices.clear();
}

unsigned int Mesh::getFlags()
{
    return m_flags | m_material->getFlags();
}

void Mesh::addRectangle2D(const glm::vec2 &pos, const glm::vec2 &dim, const glm::vec2 &texCoord, const glm::vec2 &texRange, bool indexed)
{
    if(indexed)
    {
        addVertex(pos,                     texCoord);
        addVertex(pos+glm::vec2(dim.x, 0), texCoord+glm::vec2(texRange.x, 0));
        addVertex(pos+dim,                 texCoord+texRange);
        addVertex(pos+glm::vec2(0, dim.y), texCoord+glm::vec2(0, texRange.y));
        
        addTriangle(0, 1, 2);
        addTriangle(2, 3, 0);
    }
    else
    {
        addVertex(pos,                     texCoord);
        addVertex(pos+glm::vec2(dim.x, 0), texCoord+glm::vec2(texRange.x, 0));
        addVertex(pos+dim,                 texCoord+texRange);
        
        addVertex(pos+dim,                 texCoord+texRange);
        addVertex(pos+glm::vec2(0, dim.y), texCoord+glm::vec2(0, texRange.y));
        addVertex(pos,                     texCoord);
    }
}

void Mesh::setIsDoubleSided(bool val)
{
    m_isDoubleSided = val;
    if(val)
        m_flags |= 1 << MESH_DOUBLE_SIDED;
    else
        m_flags &= ~(1 << MESH_DOUBLE_SIDED);
}

void Mesh::setIsBillboard(bool val)
{
    m_isBillboard = val;
    if(val)
        m_flags |= 1 << MESH_BILLBOARD;
    else
        m_flags &= ~(1 << MESH_BILLBOARD);
}

void Mesh::setWireframe(bool val)
{
    m_isWireframe = val;
    if(val)
        m_flags |= 1 << MESH_WIREFRAME;
    else
        m_flags &= ~(1 << MESH_WIREFRAME);
}

void Mesh::setIsShadowCaster(bool val)
{
    m_isShadowCaster = val;
    if(val)
        m_flags |= 1 << MESH_SHADOWED;
    else
        m_flags &= ~(1 << MESH_SHADOWED);
}

struct VertexComparator
{
    // c'est plutot crade mais j'ai pas trouve d'autre moyen pour le moment
    static Mesh* mesh;
    static void setMesh(Mesh* m) {VertexComparator::mesh = m;}

    bool operator() (const int& vertId1, const int& vertId2) const
    {
        if(mesh->m_positions3D[vertId1].x != mesh->m_positions3D[vertId2].x)
            return (mesh->m_positions3D[vertId1].x < mesh->m_positions3D[vertId2].x);
        if(mesh->m_positions3D[vertId1].y != mesh->m_positions3D[vertId2].y)
            return (mesh->m_positions3D[vertId1].y < mesh->m_positions3D[vertId2].y);
        if(mesh->m_positions3D[vertId1].z != mesh->m_positions3D[vertId2].z)
            return (mesh->m_positions3D[vertId1].z < mesh->m_positions3D[vertId2].z);
        if(!mesh->m_texCoords.empty())
        {
            if(mesh->m_texCoords[vertId1].x - floor(mesh->m_texCoords[vertId1].x) != mesh->m_texCoords[vertId2].x - floor(mesh->m_texCoords[vertId2].x))
                return (mesh->m_texCoords[vertId1].x < mesh->m_texCoords[vertId2].x);
            if(mesh->m_texCoords[vertId1].y - floor(mesh->m_texCoords[vertId1].y) != mesh->m_texCoords[vertId2].y - floor(mesh->m_texCoords[vertId2].y))
                return (mesh->m_texCoords[vertId1].y < mesh->m_texCoords[vertId2].y);
        }
        if(!mesh->m_normals.empty())
        {
            if(mesh->m_normals[vertId1].x != mesh->m_normals[vertId2].x)
                return (mesh->m_normals[vertId1].x < mesh->m_normals[vertId2].x);
            if(mesh->m_normals[vertId1].y != mesh->m_normals[vertId2].y)
                return (mesh->m_normals[vertId1].y < mesh->m_normals[vertId2].y);
            if(mesh->m_normals[vertId1].z != mesh->m_normals[vertId2].z)
                return (mesh->m_normals[vertId1].z < mesh->m_normals[vertId2].z);
        }
        return false;
    }
};

Mesh* VertexComparator::mesh = NULL;

void Mesh::mergeVertices()
{
    if(m_positions3D.empty())
        return;
    bool *deleted = new bool[m_positions3D.size()];
    int *offsets = new int[m_positions3D.size()];
    std::set<int, VertexComparator> vertexSet;
    VertexComparator::setMesh(this);

    for(std::size_t i=0; i<m_indices.size(); ++i)
    {
        std::pair<std::set<int,VertexComparator>::iterator, bool> ret = vertexSet.insert(m_indices[i]);
        std::set<int,VertexComparator>::iterator it = ret.first;
        bool success = ret.second;
        deleted[m_indices[i]] = !success && *it != int(m_indices[i]);
        if(deleted[m_indices[i]])
        {
            if(!m_tangents.empty())
                m_tangents[*it].tangent += m_tangents[m_indices[i]].tangent;
            m_indices[i] = *it;
        }
    }
    int offset = 0;
    int pos = 0;
    for(std::size_t i=0; i<m_positions3D.size(); ++i)
    {
        if(deleted[i])
            ++offset;
        else
        {
            offsets[i] = offset;
            if(offset != 0)
            {
                m_positions3D[pos] = m_positions3D[i];
                if(!m_texCoords.empty())
                    m_texCoords[pos] = m_texCoords[i];
                if(!m_normals.empty())
                    m_normals[pos] = m_normals[i];
                if(!m_tangents.empty())
                    m_tangents[pos] = m_tangents[i];
            }
            ++pos;
        }
    }
    for(std::size_t i=0; i<m_indices.size(); ++i)
        m_indices[i] -= offsets[m_indices[i]];

    m_positions3D.resize(m_positions3D.size()-offset);
    if(!m_texCoords.empty())
        m_texCoords.resize(m_texCoords.size()-offset);
    if(!m_normals.empty())
        m_normals.resize(m_normals.size()-offset);
    if(!m_tangents.empty())
        m_tangents.resize(m_tangents.size()-offset);
    for(int i=0; i<int(m_tangents.size()); ++i)
    {
        glm::vec3 &T = m_tangents[i].tangent;
        const glm::vec3 &N = m_normals[i];
        
        T = glm::normalize(T);
        // re-orthogonalize T with respect to N
        T = glm::normalize(T - glm::dot(T, N) * N);
        // then retrieve perpendicular vector B with the cross product of T and N
        m_tangents[i].binormal = glm::normalize(glm::cross(T, N));
    }
}

void Mesh::computeNeighbors()
{
    if(m_positions3D.empty())
        return;
    // TODO : compute adjacency and change primitivetype
}

void Mesh::computeNormals()
{
    if(m_positions3D.empty())
        return;
    m_normals.resize(m_positions3D.size());
    std::memset(m_normals.data(), 0, m_normals.size());
    for (std::size_t i=0; i < m_indices.size(); i += 3)
    {
        int v0 = m_indices[i];
        int v1 = m_indices[i+1];
        int v2 = m_indices[i+2];
        glm::vec3 n = glm::cross(m_positions3D[v1] - m_positions3D[v0], m_positions3D[v2] - m_positions3D[v0]);
        m_normals[v0] += n;
        m_normals[v1] += n;
        m_normals[v2] += n;
    }
    for(glm::vec3 &n : m_normals)
        n = glm::normalize(n);
}

void Mesh::computeTangents()
{
    if(m_texCoords.empty())
        return;
    m_tangents = std::vector<Tangents>(m_positions3D.size());

    for (std::size_t j=0; j < m_indices.size(); j += 3)
    {
        int vertexId0 = m_indices[j];
        int vertexId1 = m_indices[j+1];
        int vertexId2 = m_indices[j+2];

        const glm::vec3 &v1 = m_positions3D[vertexId0];
        glm::vec3 edge1 = m_positions3D[vertexId1] - v1;
        glm::vec3 edge2 = m_positions3D[vertexId2] - v1;
        
        const glm::vec2& w1 = m_texCoords[vertexId0];
        glm::vec2 deltaUV1 = m_texCoords[vertexId1] - w1;
        glm::vec2 deltaUV2 = m_texCoords[vertexId2] - w1;
        
        float f = 1.0f / (deltaUV1.x * deltaUV2.y - deltaUV2.x * deltaUV1.y);
        
        glm::vec3 tangentDir = deltaUV2.y * edge1 - deltaUV1.y * edge2;
        tangentDir = glm::normalize(tangentDir*f);
        
        glm::vec3 binormalDir = deltaUV1.x * edge2 - deltaUV2.x * edge1;
        binormalDir = glm::normalize(binormalDir*f);
        if(glm::dot(glm::cross(m_normals[vertexId0], binormalDir), tangentDir) < 0.f)
            binormalDir = -binormalDir;
        
        m_tangents[vertexId0] = {tangentDir, binormalDir};
        m_tangents[vertexId1] = {tangentDir, binormalDir};
        m_tangents[vertexId2] = {tangentDir, binormalDir};
    }
}

void Mesh::computeBoundingBox(glm::vec3 &min, glm::vec3 &max)
{
    min = max = m_positions3D[0];
    for(const glm::vec3 &pos : m_positions3D)
    {
        if(pos.x < min.x)
            min.x = pos.x;
        else if(pos.x > max.x)
            max.x = pos.x;
        if(pos.y < min.y)
            min.y = pos.y;
        else if(pos.y > max.y)
            max.y = pos.y;
        if(pos.z < min.z)
            min.z = pos.z;
        else if(pos.z > max.z)
            max.z = pos.z;
    }
}
