#ifndef MESH_H
#define MESH_H

#include "opengl.h"
#include <vector>
#include <string>
#include <glm/vec3.hpp>
#include <glm/vec2.hpp>

#include <SparrowSerializer/serializable.h>

class Buffer;
class Material;
class Shader;

struct Mesh : public Serializable
{
public:

    /*************************************************************/
    /*                   MESH AND MATERIAL FLAGS                 */
    /*************************************************************/

    // Mesh properties
    enum {
        // Geometric properties
        MESH_INDEXED,
        MESH_TEXTURABLE,
        MESH_INSTANCED,

        // Mesh type
        MESH_3D,
        MESH_2D,

        // 3D Geometric properties
        MESH_TANGENT_SPACE,
        MESH_DOUBLE_SIDED,
        MESH_BILLBOARD,
        MESH_WIREFRAME,
        MESH_SHADOWED,

        // simple material (no lighting)
        MATERIAL_COLOR_TEXTURE,
        MATERIAL_ALPHA_MASK,

        // 3D PBR material
        MATERIAL_PBR,
        MATERIAL_PBR_ALBEDO_TEXTURE,
        MATERIAL_PBR_EMISSION_TEXTURE,
        MATERIAL_PBR_ROUGHNESS_TEXTURE,
        MATERIAL_PBR_METALLIC_TEXTURE,
        MATERIAL_PBR_NORMAL_MAP,

        // 3D Beckman-like materials
        MATERIAL_BECKMANN_BUMP_MAP,

        NB_FLAGS
    };

    // shader "#define" strings associated to the properties
    static const char* const flagStr[NB_FLAGS];

    /*************************************************************/
    /*                      GEOMETRY DATA                        */
    /*************************************************************/

    // 3D public data
    P_VECTOR_VEC3(m_positions3D)
    P_VECTOR_VEC3(m_normals)
    
    typedef struct
    {
        glm::vec3 tangent;
        glm::vec3 binormal;
    } Tangents;

    std::vector<Tangents> m_tangents;
    
    // 2D public data
    P_VECTOR_VEC2(m_positions2D)

    // public data common to 2D and 3D
    P_VECTOR_VEC2(m_texCoords)
    P_VECTOR_VEC3(m_instances_offsets)
    P_VECTOR_UINT(m_indices)
    
    /*************************************************************/
    /*                        MAIN METHODS                       */
    /*************************************************************/

    SERIALIZABLE(Mesh,
            CAST(m_positions3D),
            CAST(m_normals),
            CAST(m_positions2D),
            CAST(m_texCoords),
            CAST(m_instances_offsets),
            CAST(m_indices),
            CAST(m_name),
            CAST(m_isDoubleSided),
            CAST(m_isBillboard),
            CAST(m_isWireframe),
            CAST(m_isShadowCaster),
            CAST(m_depth),
            CAST(m_primitive_type))

    /**
     * @brief Mesh builds an empty mesh, to be renderable, a mesh must have vertices and a material.
     */
    Mesh();
    Mesh(const std::string &name);
	~Mesh();
    
    void setName(const std::string &name) { m_name = name; }
    std::string getName() { return m_name; }
    
    /**
     * OpenGL methods :
     * - initGL allocates the GPU buffers and sends the mesh data to the GPU
     * - draw binds the material attributes and draws the mesh geometry
     * - destroyGL releases the allocated GPU memory, destroyGL is called automatically in Mesh's destructor and at the beginning of initGL
     */
    void initGL();
    void draw(Shader* shader);
    void destroyGL();
    
    /**
     * @brief clearData clears all data vectors, but keeps GPU data allocated.
     */
    void clearData();

    /**
     * @brief getFlags returns the flags that defines the specificities of the mesh and his material.
     * @warning the flags are correct only if initGL has been called.
     */
    unsigned int getFlags();

    /*************************************************************/
    /*                   MESH BUILDING METHODS                   */
    /*************************************************************/

    // add vertex
    void addVertex(float x, float y, float z) {addVertex(glm::vec3(x, y, z));}
    void addVertex(const glm::vec3 &position) {m_positions3D.push_back(position);}
    void addVertex(float x, float y) {addVertex(glm::vec2(x, y));}
    void addVertex(const glm::vec2 &position) {m_positions2D.push_back(position);}
    void addVertex(const glm::vec3 &position, const glm::vec2 &texCoord) {addVertex(position); addTexCoord(texCoord);}
    void addVertex(const glm::vec2 &position, const glm::vec2 &texCoord) {addVertex(position); addTexCoord(texCoord);}
    void addVertex(const glm::vec3 &position, const glm::vec3 &normal) {addVertex(position); addNormal(normal);}
    void addVertex(const glm::vec3 &position, const glm::vec3 &normal, const glm::vec2 &texCoord) {addVertex(position, normal); addTexCoord(texCoord);}
    void addRectangle2D(const glm::vec2 &pos, const glm::vec2 &dim, const glm::vec2 &texCoord, const glm::vec2 &texRange, bool indexed = false);
    void addRectangle2D(const glm::vec2 &pos, const glm::vec2 &dim, bool indexed = false) {addRectangle2D(pos, dim, glm::vec2(0), glm::vec2(1), indexed);}
    void addRectangle2D(float x, float y, float width, float height, bool indexed = false) {addRectangle2D(glm::vec2(x, y), glm::vec2(width, height), indexed);}
    
    // add indices
    void addTriangle(int i1, int i2, int i3) {m_indices.push_back(i1), m_indices.push_back(i2), m_indices.push_back(i3);}
    void addLine(int i1, int i2) {m_indices.push_back(i1), m_indices.push_back(i2);}

    // Material accessers
    void setMaterial(Material* mat) {m_material = mat;}
    Material *getMaterial() {return m_material;}

    // other accessers
    void addNormal(float x, float y, float z) {addNormal(glm::vec3(x, y, z));}
    void addNormal(const glm::vec3 &normal) {m_normals.push_back(normal);}
    void addTexCoord(float u, float v) {addTexCoord(glm::vec2(u, v));}
    void addTexCoord(const glm::vec2 &texCoord) {m_texCoords.push_back(texCoord);}

    /*************************************************************/
    /*                    2D MESH PROPERTIES                     */
    /*************************************************************/

    /**
     * @brief setDepth allows to set the depth of a 2D mesh, the depth must be between -1 and 1, -1 being the closest to the camera
     */
    void setDepth(float d) { m_depth = d; }
    float getDepth() const { return m_depth; }

    /*************************************************************/
    /*                    3D MESH PROPERTIES                     */
    /*************************************************************/

    /**
     * @brief setIsDoubleSided allows to enable or disable face culling for this Mesh
     */
    void setIsDoubleSided(bool val);

    /**
     * @brief setIsBillboard allows to enable or disable billboarding,
     * a billboard mesh will always follow the camera orientation
     */
    void setIsBillboard(bool val);

    /**
     * @brief setIsWireframe must be called for a mesh of GL_LINE primitives,
     * lighting will probably not be computed on lines,
     * normals value can be interpreted as a color, depending on the shader.
     */
    void setWireframe(bool val);
    
    /**
     * @brief setIsShadowCaster allows to enable or disable,
     * rendering of this mesh in the shadowmaps
     */
    void setIsShadowCaster(bool val);

    /*************************************************************/
    /*                      3D MESH TOOLS                        */
    /*************************************************************/

    /**
     * @brief mergeVertices simplifies an indexed mesh by merging all vertices that
     * have exactly the same position, texcoord, and normals, the tangents will be averaged.
     */
    void mergeVertices();

    /**
     * @brief computeNeighbors computes adjacency for a triangle mesh, the mesh type changes to GL_TRIANGLES_ADJACENCY
     */
    void computeNeighbors();

    /**
     * @brief computeNormals computes normals from an indexed mesh (positions + indices)
     */
    void computeNormals();

    /**
     * @brief computeTangents computes tangent space from a textured indexed mesh (positions + normals + texcoords + indices)
     */
    void computeTangents();
    
    /**
     * @brief computeBoundingBox computes the bounding box of the mesh based on the 3D positions
     */
    void computeBoundingBox(glm::vec3 &min, glm::vec3 &max);

    /*************************************************************/
    /*                   ADVANCED CUSTOMISATION                  */
    /*************************************************************/

    /**
     * Specifies what kind of primitives to render. Symbolic constants GL_POINTS, GL_LINE_STRIP, GL_LINE_LOOP, GL_LINES, GL_LINE_STRIP_ADJACENCY,
     * GL_LINES_ADJACENCY, GL_TRIANGLE_STRIP, GL_TRIANGLE_FAN, GL_TRIANGLES, GL_TRIANGLE_STRIP_ADJACENCY, GL_TRIANGLES_ADJACENCY and GL_PATCHES are accepted.
     *
     * default is GL_TRIANGLES
     */
    void setPrimitiveType(GLenum type) {m_primitive_type = type;}
    
    /**
     * @brief updateFlags forces the refresh of the mesh flags,
     * they will be computed from the CPU-side buffers status
     * (so don't call it after a call to clearData)
     */
    unsigned int updateFlags();
    
protected:
    P_STRING(m_name)
    
    Material* m_material;
    P_BOOL(m_isDoubleSided)
    P_BOOL(m_isBillboard)
    P_BOOL(m_isWireframe)
    P_BOOL(m_isShadowCaster)
    P_FLOAT(m_depth)
    unsigned int m_flags;

    P_UINT(m_primitive_type)

    std::vector<Buffer*> buffers;

    enum {
        // required buffer
        POSITION2D_BUFFER,
        POSITION3D_BUFFER,

        // indices buffers
        INDICES_BUFFER,

        // optionnal buffers :
        NORMAL_BUFFER,
        TEXCOORD_BUFFER,
        TANGENT_BUFFER,

        // instanciation buffer
        INSTANCE_BUFFER,

        NB_BUFFERS
    };

    int buffersId[NB_BUFFERS];

    void addBuffer(Buffer *b, int bufferType);
    void clearBuffers();

    GLuint m_vao;
};

#endif // MESH_H
