#ifndef SCENE_H
#define SCENE_H

#include <glm/mat4x4.hpp>
#include <vector>
#include "light.h"

class Camera;
class Pipeline;
class Mesh;

// Scene interface :

template <class T>
class SceneIterator
{
public:
    virtual SceneIterator& operator++() = 0;
    virtual T operator*() = 0;
    virtual bool isValid() = 0;
    void next() {if(isValid()) operator++();}
    T getItem() {return operator*();}
};

struct GeometryNode
{
    Mesh* mesh;
    glm::mat4 modelMatrix;
    
    GeometryNode(Mesh* m, glm::mat4 transform) : mesh(m), modelMatrix(transform) {}
};

class Scene
{
protected:
    Pipeline *m_pipeline;

public:
    Scene() : m_pipeline(NULL) {}

    Pipeline* getPipeline() {return m_pipeline;}

    virtual SceneIterator<Light*>* getLights() = 0;
    virtual SceneIterator<GeometryNode*>* getGeometry() = 0;

    virtual void getMeshTypes(std::vector<unsigned int> &meshTypes);
    virtual void getLightTypes(std::vector<unsigned int> &lightTypes);
};

// A basic implementation :

template <class T>
class ArrayIterator : public SceneIterator<T>
{
    std::vector<T> &vec;
    unsigned int id;
public:
    ArrayIterator(std::vector<T> &myVec, int myId=0) : vec(myVec), id(myId) {}
    virtual SceneIterator<T>& operator++() {++id; return *this;}
    virtual T operator*() {return vec[id];}
    virtual bool isValid() {return id < vec.size();}
};

class BasicScene : public Scene
{
protected:
    std::vector<Light*> lights; // fast node access for rendering
    std::vector<GeometryNode*> geometry;

    // this is used when geometrynodes are added AND allocated by the scene (with addMesh(Mesh* m, glm::mat4 transform))
    std::vector<GeometryNode*> allocations;
public:
    BasicScene() : Scene() {}
    ~BasicScene();

    void setPipeline(Pipeline *pipeline) {m_pipeline = pipeline;}
    void clearScene();
    void addMesh(GeometryNode* node);
    GeometryNode* addMesh(Mesh* m, glm::mat4 transform);
    void addLight(Light* myLight);
    Mesh* getMesh(int id) {return geometry[id]->mesh;}

    SceneIterator<Light*>* getLights();
    SceneIterator<GeometryNode*>* getGeometry();
};

#endif // SCENE_H
